!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2009  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief Calculate stress tensor and pressure numerically
!> \par History
!>      10.2005 created [JCS]
!> \author JCS
! *****************************************************************************
MODULE qs_numerical_pressure
  USE cell_types,                      ONLY: cell_clone,&
                                             cell_create,&
                                             cell_release,&
                                             cell_type,&
                                             init_cell,&
                                             real_to_scaled,&
                                             scaled_to_real
  USE cp_para_types,                   ONLY: cp_para_env_type
  USE cp_subsys_types,                 ONLY: cp_subsys_type
  USE f77_blas
  USE force_env_types,                 ONLY: force_env_get,&
                                             force_env_type
  USE global_types,                    ONLY: global_environment_type
  USE input_constants,                 ONLY: debug_run
  USE kinds,                           ONLY: dp
  USE particle_types,                  ONLY: particle_type
  USE qs_energy,                       ONLY: qs_energies
  USE qs_environment_methods,          ONLY: qs_env_rebuild_pw_env
  USE qs_environment_types,            ONLY: set_qs_env
  USE qs_ks_methods,                   ONLY: qs_ks_create
  USE qs_ks_types,                     ONLY: qs_ks_env_type,&
                                             qs_ks_release
  USE virial_types,                    ONLY: virial_type
#include "cp_common_uses.h"

  IMPLICIT NONE
  PRIVATE
  CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_numerical_pressure'

  PUBLIC :: qs_calc_numerical_pressure

  CONTAINS

! *****************************************************************************
  SUBROUTINE qs_calc_numerical_pressure(force_env, error)
    TYPE(force_env_type), POINTER            :: force_env
    TYPE(cp_error_type), INTENT(inout)       :: error

    CHARACTER(len=*), PARAMETER :: routineN = 'qs_calc_numerical_pressure', &
      routineP = moduleN//':'//routineN
    REAL(kind=dp), PARAMETER                 :: Dx = 0.001_dp

    INTEGER                                  :: i, ip, iq, iseq, iw, j, k, &
                                                stat
    LOGICAL                                  :: failure
    REAL(kind=dp), DIMENSION(2)              :: numer_energy
    REAL(kind=dp), DIMENSION(3)              :: s
    REAL(kind=dp), DIMENSION(3, 3)           :: numer_stress = 0.0_dp
    REAL(kind=dp), DIMENSION(:, :), POINTER  :: pos_ref
    TYPE(cell_type), POINTER                 :: cell, cell_local, cell_ref
    TYPE(cp_logger_type), POINTER            :: logger
    TYPE(cp_para_env_type), POINTER          :: para_env
    TYPE(cp_subsys_type), POINTER            :: subsys
    TYPE(global_environment_type), POINTER   :: globenv
    TYPE(particle_type), DIMENSION(:), &
      POINTER                                :: particles
    TYPE(qs_ks_env_type), POINTER            :: new_ks_env
    TYPE(virial_type), POINTER               :: virial

!----------------------------------------------------------------------------

    NULLIFY(cell_local)
    failure = .FALSE.
    logger=>cp_error_get_logger(error)
    IF (.NOT.failure) THEN
       !
       ! Copy atomic positions into pos_ref
       !
       CALL force_env_get(force_env,subsys=subsys,globenv=globenv,error=error)
       para_env=>force_env%para_env
       iw = cp_logger_get_default_unit_nr(logger)

       IF (iw>0) THEN
       WRITE ( iw, '( /A,A/ )' ) ' **************************** ', &
           'NUMERICAL STRESS ********************************'
       END IF

       iseq = subsys%particles%n_els
       ALLOCATE( pos_ref(iseq,3), stat=stat)
       CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
       iseq = 0
       particles => subsys%particles%els
       DO ip = 1, subsys%particles%n_els
          iseq = iseq + 1
          pos_ref(iseq,:) = particles(ip)%r
       END DO
       CALL force_env_get(force_env,cell=cell,cell_ref=cell_ref,&
                          virial=virial, error=error)
       CALL cell_create(cell_local,error=error)
       CALL cell_clone(cell,cell_local,error=error)
       !
       ! First change box.
       !
       DO ip = 1, 3
          DO iq = 1, 3
             IF ( virial%pv_diagonal .AND. ip /= iq ) CYCLE
             DO k = 1, 2
                cell%hmat(ip,iq) = cell_local%hmat(ip,iq)  - (-1.0_dp)**k * Dx
                CALL init_cell( cell )
                iseq = 0
                !
                ! Then scale positions.
                !
                particles => subsys%particles%els
                DO j = 1, subsys%particles%n_els
                   iseq = iseq + 1
                   CALL real_to_scaled(s,pos_ref(iseq,1:3),cell_local)
                   CALL scaled_to_real(particles(j)%r,s,cell)
                END DO
                !
                ! Since the box has changed, rebuild grids, i.e. pw_env and ks_env
                !
                NULLIFY ( new_ks_env )
                CALL set_qs_env(force_env%qs_env,cell=cell,error=error)
                CALL qs_env_rebuild_pw_env( force_env % qs_env, error=error)
                CALL qs_ks_create ( new_ks_env, force_env % qs_env ,error=error)
                CALL set_qs_env(force_env % qs_env, ks_env=new_ks_env ,error=error)
                CALL qs_ks_release ( new_ks_env ,error=error)
                !
                ! Compute energies
                !
                CALL qs_energies(force_env%qs_env,globenv,&
                                 consistent_energies=.TRUE.,error=error)
                numer_energy(k) = force_env%qs_env%energy%total
                ! reset cell
                cell%hmat(ip,iq) = cell_local%hmat(ip,iq)
             END DO
             CALL init_cell( cell )
             numer_stress(ip,iq) = (numer_energy(1) - numer_energy(2) ) / (2.0_dp*Dx)
          END DO
       END DO
       !
       ! Reset postions and rebuild environment.
       !
       iseq = 0
       particles => subsys%particles%els
       DO ip = 1, subsys%particles%n_els
          iseq = iseq + 1
          particles(ip)%r = pos_ref(iseq,:)
       END DO
       CALL init_cell ( cell )

       ! computing pv_test
       virial%pv_virial = 0.0_dp
       DO i = 1, 3
          DO j = 1, 3
             DO k = 1, 3
                                                              ! is the -1 a QS convention issue ?
                virial%pv_virial(i,j) = virial%pv_virial(i,j) + (-1.0_dp)*numer_stress(i,k)*cell_local % hmat(j,k)
             END DO
          END DO
       END DO
       IF (globenv%run_type_id.EQ.debug_run) THEN
           IF (para_env%ionode) THEN
               WRITE (UNIT=iw,FMT="((T34,A))") "Numerical pv_virial"
               WRITE (UNIT=iw,FMT="((T34,3F16.10))") (virial%pv_virial(i,1:3), i=1,3)
           END IF
       END IF

       !
       ! Rebuild the appropriate environments to obtain original environment
       !
       NULLIFY ( new_ks_env )
       CALL set_qs_env(force_env%qs_env,&
                       cell=cell,error=error)
       CALL qs_env_rebuild_pw_env( force_env % qs_env, error=error)
       CALL qs_ks_create ( new_ks_env, force_env % qs_env ,error=error)
       CALL set_qs_env (force_env % qs_env, ks_env=new_ks_env ,error=error)
       CALL qs_ks_release ( new_ks_env ,error=error)

       IF (para_env%ionode) THEN
          WRITE ( iw, '( /A,A/ )' ) ' **************************** ', &
           'NUMERICAL STRESS END *****************************'
       END IF

       IF (ASSOCIATED( pos_ref )) THEN
         DEALLOCATE( pos_ref, stat=stat )
         CPPostcondition(stat==0,cp_failure_level,routineP,error,failure)
       END IF
       IF (ASSOCIATED( cell_local )) CALL cell_release( cell_local ,error=error)
    END IF

  END SUBROUTINE qs_calc_numerical_pressure

END MODULE qs_numerical_pressure

